from multiclass_classification import MulticlassClassification
from preprocessing import create_emoticons_tags, create_url_tags, create_time_tags, clean_texts, \
                                       stemming, oversampling, w2v, pipeline



#!/usr/bin/python

""" Pre-processing data to do Sentiment Analysis & Multiclass Classification of portuguese Tweets. """

# Imports & Setup.
import pandas as pd
from nltk.stem import RSLPStemmer
import re
import string
import gensim
import emoticons


def create_emoticons_tags(column: pd.Series):
    """ Return texts with tags on emoticons. """

    emoticon = emoticons.emoticons_dict
    dict_emoticons = {}
    for key in emoticon.keys():
        for value in emoticon[key]:
            dict_emoticons.update({value: key})

    tagged_texts = []
    for item in column:
        for emj in dict_emoticons.keys():
            if emj in item:
                item = re.sub(emj, dict_emoticons[emj], item)
        tagged_texts.append(item)
    return tagged_texts


def create_url_tags(column: pd.Series):
    """
        Return texts with tags on URLs
        http://www.link.com -> "URL"
    """

    for index, item in enumerate(column):
        for word in item.split():
            if re.match('(https?://(?:www\.|(?!www))[a-zA-Z0-9][a-zA-Z0-9-]+[a-zA-Z0-9]\.[^\s]{2,}|www\.[a-zA-Z0-9][a-zA-Z0-9-]+[a-zA-Z0-9]\.[^\s]{2,}|https?://(?:www\.|(?!www))[a-zA-Z0-9]+\.[^\s]{2,}|www\.[a-zA-Z0-9]+\.[^\s]{2,})', word):
                item = item.replace(word, 'URL')

    return pd.Series(column)


def create_time_tags(column: pd.Series):
    """
        Return texts with tags on time
        13:30 -> 'horário'
        13h -> 'hora'
        12 min -> 'minuto'
        and so on..
    """
    clear_column = []
    for text in column:
        clear_text = []
        for word in text.split():
            word = re.sub('[1-9]hora|[0-5][0-9]horas|[0-5][0-9]h|[0-9]h', 'hora', word)
            word = re.sub('[1-9]minuto|[0-5][0-9]minutos|[0-5][0-9]min|[0-9]min|[0-5][0-9]m|[0-9]m', 'minuto', word)
            word = re.sub("[1-9]segundo|[0-5][0-9]segundos|[0-5][0-9]seg|[0-9]seg|[0-5][0-9]s|[0-9]s", 'segundo', word)
            word = re.sub('(?:[01]\d|2[0-3]):[0-5][0-9]', 'horário', word)

            clear_text.append(word.lower())
        clear_column.append(' '.join(clear_text))
    return clear_column


def clean_texts(column: pd.Series, stop_words: list):
    """ Return texts free of stopwords, accents, punctuations and twitter mentions """

    string.punctuation = string.punctuation.replace('#', '')
    string.punctuation = string.punctuation.replace('@', '')

    clear_column = []
    for text in column:
        """ Removing stopwords. """
        clear_text = [word for word in str(text).split() if word not in stop_words]

        clear_words = []
        for word in clear_text:

            """ Removing Accents. """
            word = re.sub('[ÂâÃãÄä]', 'a', word)
            # word = re.sub('[Çç]', 'c', word)
            word = re.sub('[ÈèËë]', 'e', word)
            word = re.sub('[ÌìÎîÏï]', 'i', word)
            word = re.sub('[Ññ]', 'n', word)
            word = re.sub('[ÒòÕõÖö]', 'o', word)
            word = re.sub('[Šš]', 's', word)
            word = re.sub('[ùÛÜûÙü]', 'u', word)
            word = re.sub('[ÝýŸÿ]', 'y', word)
            word = re.sub('[Žž]', 'z', word)

            """ Removing multiple vocaaaaals. """
            word = re.sub('[áàãa][áàãa]+', 'a', word.lower())
            word = re.sub('[êe][êe]+', 'e', word.lower())
            word = re.sub('[ií][ií]+', 'i', word.lower())
            word = re.sub('[oóõô][oóõô][oóõô]+', 'o', word.lower())
            word = re.sub('[uú][uú]+', 'u', word.lower())

            """ Removing @user and not A-Z chars. """
            word = re.sub('@\S+|[^A-Za-záÁãÃéÉêÊíÍóÓôÔõÕúÚçÇ]+', ' ', word)
            clear_word = ''.join([letter for letter in word if letter not in string.punctuation])

            """ Removing grouped blank spaces. """
            clear_word = re.sub('(\s)+', ' ', clear_word)

            clear_words.append(clear_word.lower())
        clear_column.append(' '.join(clear_words))

    return clear_column


def stemming(column: pd.Series, stemmer=None):
    """ Return texts stemmed. """
    if stemmer is None:
        stemmer = RSLPStemmer()
    stemmed_texts = []
    for text in column:
        stemmed_texts.append([' '.join(stemmer.stem(word)) for word in text.split()])

    return stemmed_texts


def oversampling(df: pd.DataFrame, column: str):
    """ Balance the DataFrame into equals samples of each class. """
    dict_ = {}
    for key in df[column].value_counts().keys():
        dict_.update({key: df[column].value_counts()[key]})

    sample = max(dict_.values())

    df_oversampled = pd.concat(
        [df[df[column] == key].sample(sample, replace=True, random_state=2) for key in dict_.keys()])

    return df_oversampled


def w2v(df: pd.DataFrame, column: str, max_len: int):
    """ Creates and train word2vec model. """

    w2v_model = gensim.models.word2vec.Word2Vec(size=max_len, window=7, min_count=10, workers=8, seed=5)

    """ Splitting all texts into words. """
    tweets = [text.split() for text in df[column]]

    """ Building the vocabulary. """
    w2v_model.build_vocab(tweets)

    """ Getting size of vocab. """
    # words = w2v_model.wv.vocab.keys()
    # vocab_size = len(words)
    # print("Vocab size: ", vocab_size)

    """ Training the w2v model. """
    w2v_model.train(tweets, total_examples=len(tweets), epochs=32)

    return w2v_model


def pipeline(df: pd.DataFrame, column: str, stop_words):
    """ Executes all pre-processing methods. """

    df[column] = create_emoticons_tags(df[column])
    df[column] = create_url_tags(df[column])
    df[column] = create_time_tags(df[column])
    df[column] = clean_texts(df[column], stop_words)

    return df



#!/usr/bin/python

# Imports & Setup.
import numpy as np
import pandas as pd
import pickle
import preprocessing as pp

from keras.callbacks import ReduceLROnPlateau, EarlyStopping
from keras.layers import Embedding
from keras.preprocessing.sequence import pad_sequences
from keras.preprocessing.text import Tokenizer
from keras.utils import np_utils
from keras.models import Sequential
from keras.layers import Dense, Dropout, LSTM

from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder
from sklearn.metrics import precision_recall_fscore_support
from sklearn.metrics import f1_score


class MulticlassClassification:

    def __init__(self, data: pd.DataFrame, text: str, target: str, output_filename: str, model=None,
                 word_embeddings=None, max_len=100, stop_words=None):
        self.data = data
        self.text = text
        self.target = target
        self.output_filename = output_filename
        self.tokenizer = Tokenizer()
        self.model = model
        self.w2v_model = word_embeddings
        self.max_len = max_len
        self.stop_words = stop_words

    @staticmethod
    def create_model(units=100, dense_neurons=3, embedding_layer=None):
        # print('creating model \n')
        model = Sequential()
        model.add(embedding_layer)
        model.add(Dropout(0.4, seed=5))
        model.add(LSTM(units, dropout=0.2, recurrent_dropout=0.2))
        model.add(Dense(dense_neurons, activation='softmax'))

        model.summary()
        return model

    def classification(self):
        """ Returns a sentiment_analysis model, trained with 'text' inputs and 'target' labels in 'data' file
            also, saves the model with the 'output filename'.
        """

        # Clean texts
        if self.stop_words is None:
            from nltk.corpus import stopwords
            self.stop_words = stopwords.words('portuguese')

        # pipeline will execute all pré-processing processes:
            # create_emoticons_tags, create_url_tags, create_time_tags, clean_texts
        self.data = pp.pipeline(self.data, column=self.text, stop_words=self.stop_words)

        # Oversampling the data to fix class imbalance
        self.data = pp.oversampling(self.data, column=self.target)

        # Creating our Word2Vec model if it not exists
        if self.w2v_model is None:
            self.w2v_model = pp.w2v(self.data, column=self.text, max_len=self.max_len)

        # Tokenizing
        self.tokenizer.fit_on_texts(self.data[self.text])
        vocab_size = len(self.tokenizer.word_index) + 1

        # Padding texts to create our input data X
        x = pad_sequences(self.tokenizer.texts_to_sequences(self.data[self.text]), maxlen=self.max_len)

        # Creating the encoder to set our Y in categorical format
        encoder = LabelEncoder()
        encoder.fit(self.data[self.target].tolist())
        y_encoded = encoder.transform(self.data[self.target].tolist())
        y = np_utils.to_categorical(y_encoded)

        # Now that we have our data preprocessed, we can split it into train/test
        x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=0.3, random_state=52)

        # Creating Embedding Matrix
        embedding_matrix = np.zeros((vocab_size, self.max_len))
        for word, i in self.tokenizer.word_index.items():
            if word in self.w2v_model.wv:
                embedding_matrix[i] = self.w2v_model.wv[word]

        # Creating Embedding Layer with Embedding Matrix
        embedding_layer = Embedding(vocab_size, self.max_len, weights=[embedding_matrix], input_length=self.max_len,
                                    trainable=False)

        # Creating the Model
        if self.model is None:
            self.model = self.create_model(
                dense_neurons=self.data[self.target].value_counts().count(),
                embedding_layer=embedding_layer
            )

            # Compile and Run the model
            self.model.compile(loss='binary_crossentropy', optimizer="adam", metrics=['accuracy'])
            callbacks = [ReduceLROnPlateau(monitor='val_loss', patience=5, cooldown=0),
                         EarlyStopping(monitor='val_accuracy', min_delta=1e-4, patience=5)]

            self.model.fit(x_train, y_train, batch_size=1024, epochs=5, validation_split=0.3,
                           verbose=1, callbacks=callbacks)

        # Evaluating
        self.make_evaluation(x_test, y_test)

        # saving model to output file
        filename = str(self.output_filename + '.sav')
        pickle.dump(self.model, open(filename, 'wb'))
        return self.model

    @staticmethod
    def ohe(score):
        """ Returns One Hot Encoding of score """
        index = score.argmax()
        vector = np.zeros(len(score), dtype=int)
        vector[index] = 1
        return vector

    def make_evaluation(self, x_test, y_test):
        ev_score = self.model.evaluate(x_test, y_test, batch_size=1024)
        # print("Accuracy = ", ev_score[1])
        # print("\nLoss = ", ev_score[0])

        f1_y = np.array([self.ohe(y) for y in self.model.predict(x_test)])
        print('F1-Score:')
        f1_score(y_test, f1_y, average='samples')

        print('Precision, Recall, F-Score, Support')
        precision_recall_fscore_support(y_test, f1_y, average='samples')

    @staticmethod
    def decode_sentiment(score, label_order: list):
        return list(dict(label_order).keys())[score.argmax()]

    def make_prediction(self, sample: str):
        # start_at = time.time()

        # Tokenize text
        x_test = pad_sequences(self.tokenizer.texts_to_sequences([sample]), maxlen=self.max_len)

        # Predict
        score = self.model.predict([x_test])[0]

        label_order = self.data[self.target].value_counts().sort_index()

        # Decode sentiment
        label = self.decode_sentiment(score, label_order)
        print('Predicted: ', label)

        return label



from distutils.core import setup
setup(
    name='fitanalytics',         # How you named your package folder (MyLib)
    packages=['fitanalytics'],   # Chose the same as "name"
    version='0.1.3',      # Start with a small number and increase it with every change you make
    license='MIT',
    description='FitAnalytics AI Library',   # Give a short description about your library
    author='Vinícius R. Ferraz',                   # Type in your name
    author_email='viniciusrfrz@gmail.com',      # Type in your E-Mail
    url='https://github.com/FitAnalytics/fitanalytics',   # Provide either the link to your github or to your website
    download_url='https://github.com/FitAnalytics/fitanalytics/archive/v_0.1.3.tar.gz',  # I explain this later on
    keywords=['fitanalytics', 'classification', 'multiclass'],   # Keywords that define your package best
    install_requires=[
        'numpy',
        'pandas',
        'nltk',
        'gensim',
        'keras',
        'tensorflow==2.1.0',
        'scikit-learn',
        'pickle-mixin',
    ],

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)
